/* ----------------------------------------------------------------------------------------------- *
 * DOOM MUS to MIDI converter - Ben Ryves 2006-2007 - http://benryves.com/ - benryves@benryves.com *
 * ----------------------------------------------------------------------------------------------- *
 * Usage should be pretty self-explanatory - MusToMidi.ConvertMusToMidi(Stream in, Stream out)     *
 * should be all you need. The code can also be used to work on arrays of bytes.                   *
 * I copied this from a C version I'd written for Chocolate DOOM, so it's not the cleanest code in *
 * the world... QueuedTime, I'm looking at you.                                                    *
 * ----------------------------------------------------------------------------------------------- *
 * Do whatever you like with the code below.                                                       *
 * ----------------------------------------------------------------------------------------------- */

using System;
using System.IO;

namespace ScetMus2Midi {

    public static class MusToMidi {

        #region MidiWriter Class

        /// <summary>Class to make writing MIDI files nice and easy.</summary>
        /// <remarks>Rather messy interface, could be tidied up a bit.</remarks>
        private class MidiWriter : BinaryWriter {
           
            public MidiWriter(Stream output)
                : base(output) {
            }

            enum MidiEventCode : byte {
                ReleaseKey = 0x80,
                PressKey = 0x90,
                AfterTouchKey = 0xA0,
                ChangeController = 0xB0,
                ChangePatch = 0xC0,
                AfterTouchChannel = 0xD0,
                PitchWheel = 0xE0,
            }

            uint queuedTime = 0; // Timestamps between sequences of MIDI events.
            public uint QueuedTime {
                get { return this.queuedTime; }
                set { this.queuedTime = value; }
            }

            long TrackSizeWriteLocation = 0; // Where to put the track length.

            /// <summary>Start writing a MIDI file.</summary>
            public void StartWriting() {
                this.Write(new byte[] {
        	        0x4D, 0x54, 0x68, 0x64, // Main header (MThd)
                    0x00, 0x00, 0x00, 0x06, // Header size
                    0x00, 0x00,             // MIDI type (0)
                    0x00, 0x01,             // Number of tracks
                    0x00, 0x46,             // Resolution (hard-coded to DOOM's rate).
                    0x4D, 0x54, 0x72, 0x6B, // Start of track (MTrk)
                    0x00, 0x00, 0x00, 0x00  // Placeholder for track length
                });
                this.queuedTime = 0;
                this.TrackSizeWriteLocation = this.BaseStream.Position - 4;
            }

            /// <summary>Finish writing a MIDI file.</summary>
            public void FinishWriting() {
                long T = this.BaseStream.Position;
                this.BaseStream.Seek(this.TrackSizeWriteLocation, SeekOrigin.Begin);
                uint TrackLength = (uint)(T - (this.TrackSizeWriteLocation + 4));
                for (int i = 0; i < 4; ++i) {
                    base.Write((byte)(TrackLength >> 24));
                    TrackLength <<= 8;
                }
                this.BaseStream.Seek(T, SeekOrigin.Begin);
            }

            /// <summary>Write a variable-length timestamp to a MIDI file.</summary>
            public void WriteTime(uint time) {
                uint buffer = time & 0x7F;

                while ((time >>= 7) != 0) {
                    buffer <<= 8;
                    buffer |= ((time & 0x7F) | 0x80);
                }

                for (; ; ) {
                    this.Write((byte)buffer);
                    if ((buffer & 0x80) != 0) {
                        buffer >>= 8;
                    } else {
                        queuedTime = 0;
                        return;
                    }
                }
            }

            /// <summary>Write the end of track marker.</summary>
            public void WriteEndTrack() {
                this.WriteTime(queuedTime);
                this.Write(new byte[] { 0xFF, 0x2F, 0x00 });
            }

            /// <summary>Write a key press event.</summary>
            public void WritePressKey(int channel, int key, int velocity) {
                this.WriteTime(queuedTime);
                this.Write((byte)((byte)MidiEventCode.PressKey | channel));
                this.Write((byte)(key  & 0x7F));
                this.Write((byte)(velocity & 0x7F));
            }

            /// <summary>Write a key release event.</summary>
            public void WriteReleaseKey(int channel, int key) {
                this.WriteTime(queuedTime);
                this.Write((byte)((byte)MidiEventCode.ReleaseKey | channel));
                this.Write((byte)(key & 0x7F));
                this.Write((byte)0x00);
            }

            /// <summary>Write a pitch wheel/bend event.</summary>
            public void WritePitchWheel(int channel, int wheel) {
                this.WriteTime(queuedTime);
                this.Write((byte)((byte)MidiEventCode.PitchWheel | channel));
                this.Write((byte)(wheel & 0x7F));
                this.Write((byte)((wheel >> 7) & 0x7F));
            }

            /// <summary>Write a patch change event.</summary>
            public void WritePatchChange(int channel, int patch) {
                this.WriteTime(queuedTime);
                this.Write((byte)((byte)MidiEventCode.ChangePatch | channel));
                this.Write((byte)(patch & 0x7F));
            }

            /// <summary>Write a valued controller change event.</summary>
            public void WriteChangeController(int channel, int control, int value) {
                this.WriteTime(queuedTime);
                this.Write((byte)((byte)MidiEventCode.ChangeController | channel));
                this.Write((byte)(control & 0x7F));
                value = Math.Min(value, 0x7F); // Quirk in vanilla DOOM's music code.
                this.Write((byte)(value & 0x7F));
            }

            /// <summary>Write a valueless controller change event.</summary>
            public void WriteChangeController(int channel, int control) {
                this.WriteChangeController(channel, control, 0);
            }
        }

        #endregion

        #region MusReader Class

        /// <summary>Class to make reading MUS files nice and easy.</summary>
        private class MusReader : BinaryReader {

            public MusReader(Stream input)
                : base(input) {

                this.Velocities = new byte[16];
                for (int i = 0; i < 16; i++) this.Velocities[i] = 127;

                this.head = new Header();
                this.head.ID = base.ReadUInt32();
                this.head.Length = base.ReadUInt16();
                this.head.Start = base.ReadUInt16();
                this.head.PrimaryChannels = base.ReadUInt16();
                this.head.SecondaryChannels = base.ReadUInt16();
                this.head.InstrumentCount = base.ReadUInt16();

                // Seek to where the data is held.
                this.BaseStream.Seek(this.Head.Start - 14, SeekOrigin.Current);

            }

            public struct Header {
                public uint ID;
                public ushort Length;
                public ushort Start;
                public ushort PrimaryChannels;
                public ushort SecondaryChannels;
                public ushort InstrumentCount;
                public bool IsValid {
                    get {
                        return this.ID == 0x1A53554D;
                    }
                }
            }

            private Header head;
            public Header Head {
                get { return this.head; }
            }

            public enum MusEventCode {
                ReleaseKey = 0x00,
                PressKey = 0x10,
                PitchWheel = 0x20,
                SystemEvent = 0x30,
                ChangeController = 0x40,
                ScoreEnd = 0x60,
            }

            // Keeps track of velocities.
            private byte[] Velocities;

            /// <summary>Read the event code and channel number from the MUS file.</summary>
            public void ReadEventAndChannel(out MusEventCode evt, out int channel, out byte[] data, out bool timestampFollows) {

                byte b = base.ReadByte();

                timestampFollows = (b & 0x80) != 0;

                evt = (MusEventCode)(b & 0x70);
                channel = b & 0x0F;
                switch (evt) {

                    case MusEventCode.ReleaseKey: // Key number.
                    case MusEventCode.PitchWheel: // Pitch wheel value.
                    case MusEventCode.SystemEvent: // Controller number.
                        data = new byte[1];
                        break;
                    
                    case MusEventCode.ChangeController: // Controller number, value.
                    case MusEventCode.PressKey:  // Key number, [velocity].
                        data = new byte[2];
                        break;

                    case MusEventCode.ScoreEnd:
                    default:
                        data = null;
                        break;
                }

                if (data != null) {
                    int dataLength = data.Length;
                    if (evt == MusEventCode.PressKey) dataLength = 1;
                    base.Read(data, 0, dataLength);
                    if (evt == MusEventCode.PressKey) {
                        if ((data[0] & 0x80) != 0) {
                            data[0] &= 0x7F;
                            data[1] = base.ReadByte();
                            Velocities[channel] = data[1];
                        } else {
                            data[1] = Velocities[channel];
                        }
                    }
                }

            }

            /// <summary>Convert a MUS channel number to a MIDI channel number.</summary>
            public static int ConvertMusChannelToMidiChannel(int channel) {
                if (channel == 15) {
                    return 9;
                } else if (channel == 9) {
                    return 15;
                } else {
                    return channel;
                }
            }

            /// <summary>Read a timestamp from the MUS file.</summary>
            public uint ReadTimestamp() {
                uint timestamp = 0;
                for (; ; ) {
                    byte b = base.ReadByte();
                    timestamp = (uint)(timestamp * 128 + (b & 0x7F));
                    if ((b & 0x80) == 0) break;
                }
                return timestamp;
            }

            /// <summary>Convert a MUS controller number to a MIDI controller number.</summary>
            public static int ConvertMusControllerToMidiController(int controller) {
                return (new int[] { 0x00, 0x20, 0x01, 0x07, 0x0A, 0x0B, 0x5B, 0x5D, 0x40, 0x43, 0x78, 0x7B, 0x7E, 0x7F, 0x79 })[controller];
            }
        }

        #endregion

        #region Main Converter

        public static void ConvertMusToMidi(Stream input, Stream output) {

            // Use the MidiWriter class to write out a MIDI file.
            MidiWriter Midi = new MidiWriter(output);
            Midi.StartWriting();

            // We'll pull data out of the MUS lump using a MusReader:
            MusReader Mus = new MusReader(input);
            bool HitScoreEnd = false;

            while (!HitScoreEnd) {

                while (!HitScoreEnd) {

                    // Storage of event data:
                    MusReader.MusEventCode Event; int Channel; byte[] EventData; bool TimestampFollows = false;

                    // Read an event:
                    Mus.ReadEventAndChannel(out Event, out Channel, out EventData, out TimestampFollows);

                    // Swap the channel around to the MIDI-friendly number:
                    Channel = MusReader.ConvertMusChannelToMidiChannel(Channel);

                    switch (Event) {
                        case MusReader.MusEventCode.ReleaseKey:
                            Midi.WriteReleaseKey(Channel, EventData[0]);
                            break;
                        case MusReader.MusEventCode.PressKey:
                            Midi.WritePressKey(Channel, EventData[0], EventData[1]);
                            break;
                        case MusReader.MusEventCode.PitchWheel:
                            Midi.WritePitchWheel(Channel, EventData[0] * 64);
                            break;
                        case MusReader.MusEventCode.SystemEvent:
                            Midi.WriteChangeController(Channel, MusReader.ConvertMusControllerToMidiController(EventData[0]));
                            break;
                        case MusReader.MusEventCode.ChangeController:
                            if (EventData[0] == 0) {
                                Midi.WritePatchChange(Channel, EventData[1]);
                            } else {
                                Midi.WriteChangeController(Channel, MusReader.ConvertMusControllerToMidiController(EventData[0]), EventData[1]);
                            }
                            break;
                        case MusReader.MusEventCode.ScoreEnd:
                            HitScoreEnd = true;
                            break;
                        default:
                            throw new Exception("Invalid MUS event code.");
                    }

                    if (TimestampFollows) break;
                }

                if (!HitScoreEnd) {
                    Midi.QueuedTime += Mus.ReadTimestamp();
                }

            }

            // Finish up!
            Midi.WriteEndTrack();
            Midi.FinishWriting();

        }

        #endregion

        #region Various extensions for working on arrays of bytes
        
        public static byte[] ConvertMusToMidi(Stream input) {
            using (MemoryStream m = new MemoryStream()) {
                MusToMidi.ConvertMusToMidi(input, m);
                return m.ToArray();
            }
        }

        public static byte[] ConvertMusToMidi(byte[] input) {
            return MusToMidi.ConvertMusToMidi(new MemoryStream(input));
        }

        #endregion
    }

}